%{

// $Author: jmartin $
// $Date: 2000/08/17 19:08:35 $
// $Source: /home/rigi/proj/cvs/Rigi/cobparse/cobparse.y,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#include <string>

#include <stdio.h>
#include <string.h>	// for strncpy and strncmp
#include <stdlib.h>	// for exit

#include "globals.h"
#include "coblex.h"

#define YYINITDEPTH 500

static int record_level = 0;
  
static string currentfile;
static string currentrecord;
static string currentlabel;
static string currentsection;
static string currentprogram;
static string *currentblock;

void yyerror( char *s )
{
    if (garbage_mode) return;

    fprintf( stderr,"error:" );
    fprintf( stderr," line %d (%d), token \"%s\": %s\n",
                lex_line, lex_column, yytext, s);
    fflush(stderr);
    if (syntax_stop) 
      if (0==strncmp(s,"syntax",6 ))
	if (!--syntax_stop) 
          exit(1);
}

void relation( const string& arc, const string& n1, const string& n2 )
{
  if (&arc && &n1 && &n2 && !quiet) {
    printf("%s\t%s\t%s", arc.c_str(), n1.c_str(), n2.c_str());
    if (write4)
      printf("\t%s,%d,%d\n", lex_file, lex_line, lex_column);
    printf("\n");
  }
}

int newprog(char *s)
{
  currentprogram = s;
  currentsection = s;
  relation("type", s, "Program");
}

int newsection( char *s )
{
  currentsection = s;
  currentlabel   = "";
  currentblock   = &currentsection;
  relation("type", s, "Section");
  if (currentprogram != "")
    relation("encloses", currentprogram, s);
  return 0;
}

int newlabel( char *s )
{
  currentlabel = s;
  currentblock = &currentlabel;
  relation("type", s, "Paragraph");
  if (currentsection != "")
    relation("encloses", currentsection, s);
  return 0;
}

int newfile( char *s )
{
  currentfile = s;
  if (currentfile != "global")
    relation("type", s, "File");
  return 0;
}

int newrecord( char *s )
{
  currentrecord = s;
  relation("type", s, "Record");
  if (currentfile != "global")
    relation("contains", currentfile, s );
  return 0;
}

int newdata( char *s )
{
  recordFields[s] = currentrecord;
  return 0;
}

string FindRecord( char *s )
{
  return recordFields[s];
}

// ------------------------------------------------------------------------
//
// $Log: cobparse.y,v $
// Revision 1.9  2000/08/17 19:08:35  jmartin
// small changes and documentation
//
// Revision 1.8  2000/08/16 21:55:57  jmartin
// lots of more improvements (more language features supported)
//
// Revision 1.7  2000/08/15 06:41:27  jmartin
// major overhaul & extensions
//
// Revision 1.6  2000/08/13 22:34:15  jmartin
// configure'able version of cobparse
//
// Revision 1.5  2000/08/13 02:27:26  jmartin
// check-in of mwhitney's changes of September 1995
//
//   Revision 1.4  1995/09/01  16:15:05  mwhitney
//   This version works on the 3 old COBOL test programs.
//   However, there are 91 S/R conflicts, from read_options containint
//   statements, dangling else, value_lists as confused with labels
//   (latter due to improper partitioning of sentence and statement)
//
//   Revision 1.3  1995/08/21  23:38:41  bcorrie
//   Added a slightly more detailed error message.
//
//   Revision 1.2  1994/04/27  22:29:59  mwhitney
//   fixed RCS commenting
//
//   Revision 1.1  1994/04/27  22:18:45  mwhitney
//   Initial revision
//
//

%}

%union{ char *name; }

%token  <name>  TID TALIT TNLIT

%token
TGARBAGE    TALIT   
TLPARENS   TRPARENS   TSEMI   TPERIOD   
TACCEPT   TACCESS  TADD   TADVANCING   
TAFTER   TALL   TALPHABETIC   
TALSO   TALTERNATE   TARE   TAREA   
TASCENDING   TASSIGN   TAT   
TAUTHOR   TAUTO   TBACKWARD   TBEFORE   
TBELL   TBLANK   TBLINK   TBLOCK   TBOTTOM   
TBY   TCALL   TCANCEL   TCHARACTER   TCHAIN
TCHARACTERS   TCLOSE   TCODE   TCODE_SET   
TCOL   TCOLLATING   TCOMMA   TCOMP   
TCOMPUTE   TCONFIG   TCONTAINS   TCOPY   
TCORR TCORRESPONDING TCURRENCY   TDATA  TDATE   TDATE_COMP   
TDATE_WRIT   TDAY   TDEC_POINT   TDECLARATIVES   TDELETE   TDELIMITED
TDEPENDING   TDESCENDING   TDICTIONARY   TDISK   
TDISPLAY   TDIVIDE   TDIVISION   TDOWN   TDUPLICATES   TDYNAMIC   
TELSE   TEND   TEND_OF_PAGE   TENVIRON   TEOP   TEQUAL   TERROR   
TESCAPE  TEXCEPTION   TEXCLUSIVE   TEXIT   TEXTEND   TFD   TFILE
TFILE_CONTROL   TFILLER   
TFIRST   TFIXED   TFOOTING   TFOR
TFROM   TFULL   TGIVING   TGO   TGREATER   
THIGH_VALUE THIGH_VALUES TI_O TI_O_CONTROL   TIDENT   
TIF   TIMMEDIATE   TIN   TINDEX   
TINDEXED   TINITIAL   TINPUT   
TINPUT_OUTPUT   TINSPECT   TINSTALLATION   
TINTO   TINVALID   TIS   TJUST   
TKEY   TKEYBOARD   TLABEL   TLEADING   
TLEFT   TLESS   TLINAGE   
TLINAGE_COUNTER   TLINE   TLINES   
TLINKAGE   TLOCK   TLOW_VALUE TLOW_VALUES  
TMEMORY   TMERGE   TMODE   
TMODULES   TMOVE   TMULT   
TMULTIPLE   TNAME   TNATIVE   
TNEGATIVE   TNEXT   TNO   TNOTAT TNOTINVALID TNUMBER   
TNUMERIC   TOBJECT_COMP   TOCCURS   
TOF   TOFF   TOMITTED   TON   TOPEN   
TOPTIONAL TORGANIZATION   TOUTPUT   
TOVERFLOW   TPAGE   TPERFORM   
TPIC   TPLUS   TPOINTER   
TPOSITION   TPOSITIVE   TPREVIOUS   
TPRINTER   TPROCEDURE   
TPROG_ID   TPROGRAM   TQUOTE   TRANDOM   
TREAD   TRECORD   TRECORDS   TRECORDING   
TREDEFINES   TRELATIVE   TRELEASE   
TREMAINDER   TRENAMES   TREPLACING   
TREQUIRED   TRERUN   TRETURN   TREWIND   
TREWRITE   TRIGHT   TROUNDED   TRUN   
TSAME   TSCREEN   TSD   
TSEARCH   TSECTION  TSECURE   TSECURITY 
TSEGMENT_LIMIT   TSELECT   
TSENTENCE   TSEPARATE   TSEQUENCE   
TSEQUENTIAL   TSET   TSIGN   
TSIZE   TSORT   TSORT_MERGE   
TSOURCE_COMP   TSPACE   TSPECIAL_NAMES   
TSTANDARD   TSTART   TSTATUS   TSTOP   
TSUBT   TSWITCH   TSYMBOLIC   TSYNC   
TTALLYING   TTHAN   TTHROUGH  TTHRU TTIME   
TTIMES   TTO   TTRAILING   TUNDELETE   
TUNLOCK   TUNSTRING   TUNTIL   TUP   
TUPON   TUSAGE   TUSE   TUSER   TUSING   
TVALUE   TVALUES   TVARIABLE   
TVARYING   TWHEN   TWITH   TWORDS   
TWRITE   TWSTOR   TZERO TZEROS   


/* Non-standard tokens */

TALLOWING TALPHANUMERIC TCONSOLE TCONTINUE TDESCRIPTOR
TEND_DIVIDE TEND_EVALUATE TEND_IF TEND_PERFORM 
TEND_READ TEND_RETURN TEND_SEARCH TEND_START TEND_STRING TEND_WRITE
TEVALUATE TFAILURE TFUNCTION TORDER TOTHER TREFERENCE TSPACES TSTRING 
TTEST TTRUE TZEROES TRETURN_CODE 

/* JCL */
TJCL

%token 
TINITIALIZE

%left 	TPLUS   TMINUS   
%left 	TSTAR   TSLASH   
%left 	TEXP   
%nonassoc 	TEQ   TLE   TLT   TGE   TGT   
%left 	TOR   
%left 	TAND   
%right 	TNOT   


%start sourcefile

%%

sourcefile      : opt_jcl programs opt_jcl 
		;

opt_jcl		: null 
		| jcl
		;

jcl		: TJCL
		| jcl TJCL
		;

programs	: program
		| programs jcl { recordFields.clear(); } program 
		;

program		: opt_id_div opt_env_div opt_data_div proc_div
		| opt_sections section
		| opt_sections paragraph
		| error TJCL { yyerror("invalid COBOL program"); }
		;


/* PRELIMINARY DIVISIONS */

opt_id_div	: null | id_div ;
opt_env_div	: null | env_div ;
opt_data_div	: null | data_div ;

id_div_hd	: TIDENT TDIVISION TPERIOD { garbage_mode = 1; } ;
id_div		: id_div_hd gstmts ;
env_div		: TENVIRON TDIVISION TPERIOD gstmts ;
data_div_hd	: TDATA TDIVISION TPERIOD
			{ garbage_mode = 0; }
		;
data_div	: data_div_hd file_section work_section linkage_section;
proc_div	: TPROCEDURE TDIVISION opt_using TPERIOD opt_decl opt_sections ;

gstmts		: gstmts gstmt | gstmt ;
gstmt		: TPROG_ID TPERIOD TID TPERIOD { newprog($3); } 
		| error TPERIOD {yyerrok;} ;

file_section	: TFILE TSECTION TPERIOD dstmts { newfile("global"); } ;
work_section	: TWSTOR TSECTION TPERIOD dstmts ;
linkage_section : null | TLINKAGE TSECTION TPERIOD dstmts ; 

dstmts		: dstmts dstmt | dstmt ;

dstmt		: dstmt_hd fd_end
		| error TPERIOD { yyerror("unrecognized data line"); yyerrok; }
		;

dstmt_hd	: TFD fd_id { garbage_mode = 1; }
		| TSD fd_id { garbage_mode = 1; }
		| TCOPY fd_id { garbage_mode = 1; }
		| level_id rc_id { garbage_mode = 1; }
		| level_id TFILLER { garbage_mode = 1; }
		| level_id { garbage_mode = 1; }
		;

fd_end		: opt_dclauses TPERIOD { garbage_mode = 0; }
		| error TPERIOD { yyerrok; garbage_mode = 0; }
		;

opt_dclauses	: null | opt_dclauses dclause ;

dclause		: TINDEXED TBY rc_id
		| TOCCURS TNLIT TTIMES
		| TOCCURS TNLIT TTIMES TDEPENDING TON TID
		| TOCCURS TNLIT TTO TNLIT TTIMES TDEPENDING TON TID
		
		;

level_id	: TNLIT	{ record_level = atoi($1); } ;
fd_id		: TID	{ newfile( $1 ); } ;

rc_id		: TID
		{
			if (record_level==1)
			{
			    newrecord( $1 );
			    newdata( $1 );
			}
			else
			{
			    newdata( $1 );
			}
		}
		;


/* PROCEDURE DIVISION */

opt_decl	: null
		| TDECLARATIVES TPERIOD
		  TID TSECTION TPERIOD
		  sentences
		  TEND TDECLARATIVES TPERIOD
		;

opt_sections	: null { garbage_mode = 0; }
		| opt_sections section
		| opt_sections paragraph
		;

paragraph	: label TPERIOD sentences ;

section		: sectname TSECTION TPERIOD sentences ;

sectname	: TID	{ newsection($1); } ;

perfname        : TID	{ relation("performs",*currentblock,$1); } ;

sentences	: sentence
		| sentences sentence
		;


sentence	: label TPERIOD
		| compound TPERIOD
		| error TPERIOD
			{ yyerror("unknown sentence"); }
		;

/*
	(compound) statements (no intervening periods) must not have
	IF statements within; only at the end so that their own
	compound statements belong to them alone (??)
	I.e., statements never follow IF statements
	(unless the latter uses the (new) END-IF token)
*/

compound	: statements
		| cond_stmt
		| statements cond_stmt
		;

statements	: statement
		| statements statement
		;

cond_stmt	: if_stmt
		| if_else_stmt
		| search_stmt
		| eval_stmt
		;

if_stmt		: TIF cond else_clause ;
if_else_stmt	: TIF cond else_clause TELSE else_clause ;

else_clause	: compound
		| TNEXT TSENTENCE
		;

search_stmt	: TSEARCH opt_all value opt_vary opt_at whens ;
eval_stmt	: TEVALUATE value whens ;

whens		: null
		| whens when
		;

when		: TWHEN cond compound
		| TWHEN TOTHER compound
		;

/* NORMAL STATEMENTS */

statement	: goto
		| move
		| perform
		| perform_loop

		| call
		| accept
		| multiply
		| divide
		| close
		| stoprun
		| compute
		| display
		| add
		| subtract
		| open
		| set
		| unstring
		| continue
		| inspect

/* Some of these have S/R conflicts from their optional clauses */

		| read
		| write
		| rewrite
		| delete
		| start
		| sort
		| merge

/* This used in declaratives only!? */

		| use

/* non-standard */

		| initialize
		| release
		| return
		| return TEND_RETURN
		| if_stmt TEND_IF
		| if_else_stmt TEND_IF
		| search_stmt TEND_SEARCH
		| eval_stmt TEND_EVALUATE 
		| read TEND_READ
		| write TEND_WRITE
		| start TEND_START
		| string
		| string TEND_STRING
		| divide TEND_DIVIDE
		| exit

		| error { yyerror("unknown statement"); }
		;

exit		: TEXIT
		| TEXIT TPROGRAM
		;

stoprun		: TSTOP TRUN ;

display		: TDISPLAY values
		| TDISPLAY values TUPON device
		;

device		: TPRINTER | TCONSOLE ;


start		: TSTART file_id TKEY opt_is opt_not rel_op val_id opt_invalid ;

delete		: TDELETE file_id TRECORD opt_invalid ;

read		: TREAD read_id opt_next opt_record opt_into opt_invalid opt_at;

write		: TWRITE write_id opt_from opt_advance opt_invalid opt_at ;

sort		: TSORT file_id opt_ons opt_dupl opt_seq sort_in sort_out ;
merge		: TMERGE file_id opt_ons opt_seq sort_in sort_out ;

opt_ons		: null
		| opt_ons opt_on;

opt_on		: TON TASCENDING TKEY values
		| TON TDESCENDING TKEY values
		| TASCENDING TKEY values
		| TDESCENDING TKEY values
		;

opt_dupl	: null
		| opt_dupl_with TDUPLICATES opt_dupl_in opt_dupl_order
		;

opt_dupl_with	: null | TWITH ;
opt_dupl_in	: null | TIN ;
opt_dupl_order	: null | TORDER ;

opt_seq		: null
		| opt_seq_coll TSEQUENCE opt_is TID
		;

opt_seq_coll	: null | TCOLLATING ;

sort_in		: TUSING fileids
		| TINPUT TPROCEDURE opt_is perfname opt_thru
		;

sort_out	: TGIVING fileids
		| TOUTPUT TPROCEDURE opt_is perfname opt_thru
		;

string		: TSTRING string_elems TINTO value ;
	
string_elems	: value opt_delim
		| string_elems value opt_delim
		;

rewrite		: TREWRITE write_id opt_from opt_invalid
		;


file_id		: TID { relation( "accesses", *currentblock, $1 ); } ;
read_id		: TID { relation( "reads", *currentblock, $1 ); } ;
write_id	: TID { relation( "writes", *currentblock, $1 ); } ;

opt_record	: null | TRECORD ;
opt_next	: null | TNEXT ;
opt_into	: null | TINTO value ;
opt_from	: null | TFROM value ;
opt_at		: null 
		| opt_at TAT TEND compound
		| opt_at TNOTAT TEND compound
		;
opt_invalid	: null 
		| opt_invalid TINVALID TKEY compound
		| opt_invalid TNOTINVALID TKEY compound
		;
opt_advance	: null 
                | TAFTER TADVANCING value TLINES
		| TAFTER TADVANCING TPAGE
                | TBEFORE TADVANCING value TLINES
      		| TBEFORE TADVANCING TPAGE
		;

open		: TOPEN file_specs
		| TOPEN file_specs TALLOWING TALL
		;

file_specs	: file_spec fileids
		| file_specs file_spec fileids
		;

fileids		: file_id | fileids file_id ;

file_spec	: TINPUT | TOUTPUT | TI_O | TEXTEND ;

close		: TCLOSE fileids ;


perform		: TPERFORM perfname opt_thru opt_with opt_vary opt_until opt_after ;
perform_loop	: TPERFORM opt_with opt_vary TUNTIL cond compound end_perform
		;

end_perform	: TEND_PERFORM 
                | error TPERIOD { yyerror("no end-perform"); }

opt_with	: null
		| TWITH TTEST TAFTER
		| TWITH TTEST TBEFORE
		| TTEST TAFTER
		| TTEST TBEFORE
		;

opt_thru	: null | thru ;
opt_vary	: null | vary ;
/*
opt_times	: null | value TTIMES ;
*/

vary		: TVARYING value TFROM value TBY value 
		| TVARYING value
		;

opt_after       : null 
		| TAFTER value TFROM value TBY value TUNTIL cond opt_after
		;

thru		: TTHROUGH perfname | TTHRU perfname ; 

accept		: TACCEPT TID 
		| TACCEPT TID TFROM value
		;

move		: TMOVE opt_all opt_corr value TTO values ;

add		: TADD opt_corr values TTO values opt_round
		| TADD opt_corr values TGIVING value opt_round
		;

subtract	: TSUBT opt_corr values TFROM values
		| TSUBT opt_corr values TFROM value TGIVING value
		;

opt_corr	: null | TCORR | TCORRESPONDING ;

multiply	: TMULT value TBY value
		| TMULT value TBY value TGIVING value opt_round
		;

divide		: TDIVIDE value TINTO value opt_round opt_remainder
		| TDIVIDE value TINTO value TGIVING value opt_round opt_remainder
		| TDIVIDE value TBY value TGIVING value opt_round opt_remainder
		;

opt_remainder	: null
		| TREMAINDER value
		;

compute		: TCOMPUTE value opt_round TEQ expr ;

inspect		: TINSPECT val_id TREPLACING TALL value TBY value ;

label		: TID   { newlabel($1); }
		;

opt_until	: null | TUNTIL cond ;
opt_round	: null | TROUNDED ;

set		: TSET values TTO value
		| TSET values TUP TBY value
		| TSET values TDOWN TBY value
		;

unstring	: TUNSTRING value opt_delim TINTO values 
		;

continue	: TCONTINUE ;

call		: TCALL callname opt_using opt_giving opt_oflow ;

callname	: TID 	{ relation("calls",*currentblock,$1); }
		| TALIT	{ relation("calls",*currentblock,$1); }
		;

opt_using	: null | TUSING opt_by uvalues;
opt_giving	: null | TGIVING value ;
opt_oflow	: null | TON TOVERFLOW compound ;
opt_delim	: null | TDELIMITED opt_by opt_all value ;
opt_all		: null | TALL ;
opt_by		: null | TBY ;

uvalue		: value | TVALUE | TDESCRIPTOR | TREFERENCE ;
uvalues		: uvalue | uvalues uvalue ;

goto		: TGO TTO TID ;


use		: TUSE TAFTER TSTANDARD TERROR TPROCEDURE TON TID
		;

return		: TRETURN file_id opt_record opt_into opt_at 
		;

release		: TRELEASE value opt_from
		;

initialize	: TINITIALIZE values 
		| TINITIALIZE values TREPLACING datatype TDATA TBY value
		;

datatype	: TALPHANUMERIC
		| TALPHABETIC
		| TNUMERIC
		;

cond		: value
		| value opt_is opt_not rel_op value
		| value opt_is opt_not rel_op TALL TSPACES
		| value opt_is opt_not datatype
		| value opt_is opt_not TPOSITIVE
		| value opt_is opt_not TNEGATIVE
		| value opt_is opt_not value

		| cond TAND cond
		| cond TAND opt_is opt_not rel_op value
		| cond TOR cond
		| cond TOR opt_is opt_not rel_op value
		| TLPARENS cond TRPARENS
		| TNOT cond
		;

ranges		: range
		| ranges range
		;

range		: expr
		| expr TCOMMA expr
		| expr ':' expr
		| TALL
		;

values		: value
		| values value
		| values TCOMMA value
		;

value		: val_id
		| val_id TLPARENS ranges TRPARENS
		| TNLIT						{ ; }
		| TALIT 					{ ; }
		| TDATE
		| TDAY
		| TTIME
		| TSPACE
		| TSPACES
		| TZERO
		| TZEROS
		| TZEROES
		| TLOW_VALUE
		| TLOW_VALUES
		| THIGH_VALUE
		| THIGH_VALUES

		| TTRUE
		| TFAILURE
		| TRETURN_CODE
		| TSIZE
		;

val_id		: TID
		{
			string s = FindRecord( $1 );
			if (s != "")
			  relation("references",*currentblock,s);
		}
		| TID TOF val_id { ; }
		;

opt_is		: null | TIS ;
opt_not		: null | TNOT ;

rel_op		: TGE | TGT | TLE | TLT | TEQ 
		| TGREATER TTHAN
		| TLESS TTHAN
		;

expr		: value
		| expr TPLUS expr
		| expr TMINUS expr
		| expr TSTAR expr
		| expr TSLASH expr
		| TLPARENS expr TRPARENS
		| TFUNCTION TID TLPARENS expr TRPARENS
		;
null		:;
%%
